""" base classes to describe python code """
__revision__ = '$Id: base.py,v 1.4 2002/01/15 12:01:47 syt Exp $'

import re

## regex to handle _ visibility
##                 _ class / abstract class / interface differentiation
##                 _ final
## in general, try to describe python code as if it was java code ;)
PROTECTED = re.compile('^_[A-Za-z]+(_*[A-Za-z]+)*_?$')
PRIVATE = re.compile('^__[A-Za-z]+(_*[A-Za-z]+)*_?$')
SPECIAL = re.compile('^__[A-Za-z]+__$')
FINAL = re.compile('^[A-Z_]*$')
ABSTRACT = re.compile('^.*Abstract.*')
INTERFACE = re.compile('^.*Interface.*')
EXCEPTION = re.compile('^.*exception.*', re.I)
CALL = re.compile('^(\w|\.)+\((\w|,)*\)')

## mixin to handle locals and globals dict ####################################
class FrameMI:
    """ Mixin to handle a dict of defined object in context
    it should be the "leftmost" ancestor
    """
    
    def __init__(self): 
        self.imports = []
        self._defs= {}
        self.raises = {} # /!\ become a list after resolve_raises /!\

    def add_import(self, fullname, asname):
        """ add an import statement to the context """
        self._defs[asname] = fullname
        self.imports.append((fullname, asname))
        
    def add_raised_exception(self, exception):
        """ add a raise statement to the context """
        self.raises[exception] = 1

    def get_ref(self, name):
        """
        asking for a reference to an object named 'name'
        if the object is not found in this context, ask to my parent
        """
#        print 'Asking for %s to %s'%(name, self.name)
        mychance = self._defs.get(name)
        if mychance and mychance != name:
#            print 'Yep!! return',repr(mychance)
            return mychance
        elif self.parent:
            subname = '.'.join(self.fullname().split('.')[:-1])
            if subname:
                other = self.parent.get_ref('%s.%s'%(subname,name))
                if other:
#                    print 'Yep2!! return',repr(other)
                    return other
            return self.parent.get_ref(name)
#        print "Carramba !!"
        
    def resolve_data_graph(self):
        """
        try to associate each value in the context dict with a reference to an
        object
        recursivly resolve children
        """
        ## resolve definitions graph
        for key, val in self._defs.items():
            if not isinstance(val, AbstractBase):
                defs = self.get_ref(val)
                if defs:
                    self._defs[key] = defs

    def resolve_imports(self):
        imp = []
        for mod, asmod in self.imports:
            if mod[-2:] == '.*':
                mod = mod[:-2]
            val = full_get_ref(self.get_ref, AbstractBase, mod)
            imp.append((val, asmod))
        self.imports = imp
        
    def resolve_exceptions(self):
        """ try to get an object reference from the name in the 'parents' list """
        raises_l = []
        if self.raises:
            for exception in self.raises.keys():
                if CALL.match(exception):
                    exception = full_get_ref(self.get_ref, Klass,
                                             exception[:exception.index('(')])
                    if isinstance(exception, AbstractBase):
                        if not hasattr(exception, 'context'):
                            exception.context = []
                        exception.context.append(self)
                raises_l.append(exception)
        self.raises = raises_l

    def appendChild(self, child):
        """ overload AbstractBase.appendChild """
        if AbstractBase.appendChild(self, child):
            self._defs[child.name] = child
            return 1

class IdGeneratorMI:
    
    def initCounter(self, start_value=0):
        self.id_count = start_value
        
    def generateId(self):
        """ generate a new identifer """
        self.id_count += 1
        return self.id_count

def depth_first(node, filter, callback):
    """
    recursivly call <callback> children which pass <filter>
    """
    if filter(node):
        callback(node)
    for child in node.childs:
        depth_first(child, filter, callback)

## classes hierarchy to describe the project implementation ###################
## base abstract classes ##

class AbstractBase:
    """ base class, all others inherit from this one """
    SP = ['childs', 'id', 'parent', 'name']

    def __init__(self, name, id):
        self.id = id
        self.name = name
        self.parent = None
        self.childs = []
        self._childs_d = {}
            
    def fullname(self):
        """ for instance return 'module_name.class_name' """
        if self.parent:
            pname = self.parent.fullname()
            if pname:
                return '%s.%s'% (pname, self.name)
        return self.name

    def root(self):
        """ return project node """
        if self.parent is None:
            return self
        return self.parent.root()

    def get_module(self):
        """ get the first ancestor which is a module """
        if self.TYPE == 'module':
            return self
        if self.parent:
            return self.parent.get_module()
        
    def get_class(self):
        """ get the first ancestor which is a class """
        if self.TYPE == 'class':
            return self
        if self.parent:
            return self.parent.get_class()
        
    def appendChild(self, child):
        """ add a new child (unique name) """
        if not self._childs_d.has_key(child.name):
            self._childs_d[child.name] = child
            self.childs.append(child)
            child.parent = self
            return 1
        return 0
    
    def __repr__(self):
        """ short representation (no recursion) """
        return '<%s name="%s" id="%d">' % (self.TYPE, self.name, self.id)

    def __str__(self, i=0):
        """ xml like representation """
        s = '%s<%s name="%s" id="%d"'%(i*'  ', self.TYPE,
                                          self.name, self.id)
        for a in self.__dict__.keys():
            if not a in self.SP:
                val = getattr(self, a)
                if val:
                    s = '%s\n%s%s="%s"'%(s,(i+1)*'  ', a, val)
        if not self.childs:
            return '%s/>\n'%s
        else:
            s = '%s>\n'%s
            for c in self.childs:
                s = '%s%s'%(s, c.__str__(i+1))
            return '%s%s</%s>\n'%(s, i*'  ', self.TYPE)

        
class AbstractFinal(AbstractBase):
    """ base class for element finalizable and with a visibility """

    def __init__(self, name, id):
        AbstractBase.__init__(self, name, id)
        if FINAL.match(name): self.final = 'true'
        else: self.final = 'false'
        if PROTECTED.match(name):
            self.visibility = 'protected'
        elif PRIVATE.match(name):
            self.visibility = 'private'
        else: self.visibility = 'public'


class AbstractAttr(AbstractFinal):
    """ base class for class child (attributes and functions) """

    def __init__(self, name, id, type=None):
        AbstractFinal.__init__(self, name, id)
        if SPECIAL.match(name):
            self.special = 1
        else: self.special = 0
        if not type is None:
            self.type = type ## instance/class

    
## classes used to describe the project ##
            
class Project(FrameMI, AbstractBase):
    """ class representing a project
    this is the root of the hierarchy tree
    a class can only contains modules
    """
    TYPE = 'project'

    def __init__(self, name, id):
        FrameMI.__init__(self)
        AbstractBase.__init__(self, name, id)

    def get_classes(self, name, module=None):
        """ return the list of classes named 'name'
        (you can optionaly specify a module name)
        """
        r = []
        if module:
            m = self.get_module(module)
            if m:
                c = module.get_class(name)
                if c: r.append(c)
        else:
            for module in self.childs:
                c = module.get_class(name)
                if c: r.append(c)
        return r

    def get_module(self, name=''):
        """ overload AbstractBase.get_module """
        if name:
            for module in self.childs:
                if module.name == name:
                    return module

    def resolve_data_graph(self):
        """ overload FrameMI.resolve_data_graph """
        ## add child modules defs to _defs:
        for child in self.childs:
            for key, val in child._defs.items():
                fname = child.fullname()
                if not self._defs.has_key('%s.%s'%(fname,key)):
                    self._defs['%s.%s'%(fname,key)] = val
                if fname != child.name:
                    self._defs[fname] = child
        ##  resolve tree graph
        for child in self.childs:
            depth_first(child, lambda x: isinstance(x, FrameMI),
                        lambda x: x.resolve_data_graph())
            
    def resolve_imports(self):
        """ overload FrameMI.resolve_imports """
        for child in self.childs:
            depth_first(child, lambda x: isinstance(x, FrameMI),
                        lambda x: x.resolve_imports())
                    
    def resolve_hierarchy(self):
        """ call resolve_hierarchy on all owned classes """
        for child in self.childs:
            depth_first(child, lambda x: isinstance(x, Klass),
                        lambda x: x.resolve_hierarchy())
                    
    def fullname(self):
        """ overload AbstractBase.fullname """
        return ''
   
class Module(FrameMI, AbstractBase):
    """ class representing a module
    a module can contain classes, attributes and functions
    """
    TYPE = 'module'
    
    def __init__(self, name, id):
        FrameMI.__init__(self)
        AbstractBase.__init__(self, name, id)

    def get_deps(self):
        if hasattr(self, '_deps'):
            return self._deps
        deps = {}
        for defi in self._defs.values():
            try:
                m = defi.get_module()
                if m.name != self.name:
                    deps[m] = 1
            except:
                pass
        self._deps = deps.keys()
        return self._deps

    def fullname(self):
        """ overload AbstractBase.fullname """
        name = self.name.split('.')
        if name[-1] == '__init__':
            return '.'.join(name[:-1])
        return '.'.join(name)

    def get_class(self, name):
        """ return the class named 'name' if present """
        for child in self.childs:
            if child.TYPE == 'class' and child.name == name:
                return child
        return None


class Klass(FrameMI, AbstractFinal):
    """ class representing a class
    a class can contain classes, attributes and functions
    """
    TYPE = 'class'
    
    def __init__(self, name, id):
        FrameMI.__init__(self)
        AbstractFinal.__init__(self, name, id)
        self.parents, self.subclasses = [], []
        if INTERFACE.match(name):
            self.type = 'interface' 
        else: self.type = 'class'
        if ABSTRACT.match(name):
            self.abstract = 'true'
        else: self.abstract= 'false'
        if EXCEPTION.match(name):
            self.signal = 'true'
        else: self.signal = 'false'

    def resolve_hierarchy(self):
        """ try to get an object reference from the name in the 'parents' list """
        if self.parents:
            parents_l = []
            for parent in self.parents:
                parents_l.append(full_get_ref(self.get_ref, Klass, parent))
                if parents_l[-1] == 'Exception':
                    self.signal = 'true'
            self.parents = parents_l
        self.resolve_all()

    def resolve_all(self):
        for child in self.childs:
            if child.TYPE == 'attribute':
                child.resolve_link()
            elif child.TYPE == 'function':
                child.resolve_exceptions()
                
class Function(FrameMI, AbstractAttr):
    """ class representing a function
    a function can contains classes, attributes and functions
    """
    TYPE = 'function'        

    def __init__(self, name, id, type=None):
        FrameMI.__init__(self)
        AbstractAttr.__init__(self, name, id, type)

    def resolve_data_graph(self):
        """ overload FrameMI.resolve_data_graph """
        ## if parent is a class, remove first parameter ?
        if self.childs and isinstance(self.parent, Klass):
            del self.childs[0]
        FrameMI.resolve_data_graph(self)
        
    def fullname(self):
        """ overload AbstractBase.fullname """
        name = self.name
        pname = self.parent.fullname()
        if pname:
            return '%s.%s'% (pname, name)
        return name


class Attribute(AbstractAttr):
    """ class representing an attribute
    an attribute can't contains any element
    """
    TYPE = 'attribute'

    def resolve_link(self):
        """ try to get an object reference from the attribute default value """
        if hasattr(self, 'default'):
            defv = self.default
            if CALL.match(defv):
                self.default = full_get_ref(self.parent.get_ref, AbstractBase,
                                            defv[:defv.index('(')])
        
class Parameter(AbstractFinal):
    """ class representing an parameter
    an parameter can't contains any element
    """
    TYPE = 'parameter'
    
    def __init__(self, name, id, type):
        AbstractFinal.__init__(self, name, id)
        self.type = type ## in/out
        
        
def full_get_ref(get_ref_func, base, name, warn=0):
    assert not isinstance(name, AbstractBase)
    refs = get_ref_func(name)
    if refs is not None and isinstance(refs, base):
        return refs
    elif refs is not None and refs!= name:
        done = [name]
        while refs and not isinstance(refs, base) and not refs in done:
            done.append(refs)
            last = refs
            refs = get_ref_func(refs)
        if refs:
            return refs
        else: return last
    if warn:
        import sys
        print >>sys.stderr, "Unresolved name %s" % name
    return name


